% Two dimention AMERICAN OPTION pricing with RBF-PUM
% exercise price E=1 domain=[0,4]x[0,4].
% PUM method is used with inverse multiquadric basis 
% for bases of local interpolants
% penalty parameters \delta=0.00001, C=r.
% Ali Safdari, Alfa Heryudono and Elisabeth Larsson - App 01 2014.
% Needed functions: 
%        weight_PUM, imq_PUM, BS_1D, PUM_BS_diffmat and odefun_2D

clc; clear all; close all; 
%AMRICAN OPTION CONSTANTS
r=0.1; sig1=0.2; sig2=0.3; alf1=.6; alf2=.4; D1=0.05; D2=0.01; ru=0; Tf = 1; 
E=1; L=0; S=4; l=E/2; Ec=0.75*E; Sc=0.75*S; lc=Ec/2;
tic
N = 21; M = 6; %number of node and patch at each axis 
% Addapted nodes generation
xs = linspace(L,S,N)'; [x,y] = meshgrid(xs,xs); data = [x(:) y(:)];
hep=min(abs(diff(xs)))*1.5;
% Create and organize centers for partitions
xce = linspace(L,S,M+1); H = min(diff(xce));
xc = xce(1:end-1) + 0.5*H; [xcen,ycen] = meshgrid(xc,xc);
center = [xcen(:) ycen(:)]; 
% Shape of the ellips
D = diff(xc); dif = [D(1) min(D(1:end-1),D(2:end)) D(end)];
[Rx, Ry] = meshgrid(dif,dif); R =.8* [Rx(:) Ry(:)];
%__________________________________________________________________________
%plot data points and centers
xcorner = [L L;S L;S S;L S;L L]; figure(1);
plot(xcorner(:,1),xcorner(:,2),'-k','linewidth',2); hold on
plot(data(:,1),data(:,2),'bo','MarkerSize',3); hold on
plot(xcen,ycen,'*r'); 
set(gca,'Fontsize',14)
set(gca,'ytick',(-0.5:0.5:S+0.5),'xtick',(-0.5:0.5:S+0.5)); 
hold on
th = linspace(0,2*pi,51);
for i=1:length(center)
    xx = center(i,1)+1i*center(i,2)+(R(i,1)*cos(th)+1i*R(i,2)*sin(th));  
    plot(xx,'-r'); hold on
end
axis equal; axis([-0.5,S+0.5,-0.5,S+0.5]); hold off; 
% %________________________________________________________________________
[Ax,Axx,Ay,Ayy,Axy] = PUM_BS_diffmat(data,center,R,hep);
% The differentiation matrices corresponding to interior nodes
ind = find(data(:,1)>min(xs) & data(:,1)<max(xs) & data(:,2)>min(xs) & data(:,2)<max(xs));
Idx = diag(data(ind,1)); Idy=diag(data(ind,2)); I=speye(length(ind),length(ind));
Wd = .5*sig1^2*Idx^2*Axx(ind,ind)+.5*ru*sig1*sig2*Idx*Idy*Axy(ind,ind)+...
     .5*sig2^2*Idy^2*Ayy(ind,ind)+.5*ru*sig2*sig1*Idx*Idy*Axy(ind,ind)+...
     -(D1-r)*Idx*Ax(ind,ind)-(D2-r)*Idy*Ay(ind,ind)-r*I;
% The near-field boundary, and the far-field boundary.
bL1 = find(data(:,1)==min(xs) & data(:,2)<max(xs));
bL2 = find(data(:,2)==min(xs) & data(:,1)<max(xs) & data(:,1)>min(xs));
bS = find(data(:,1)==max(xs) | data(:,2)==max(xs));
inb =[bL1 ;bL2 ;bS];
% Boundary values at near-field boundary by solving 1D Black-Scholes
xb1 = BS_1D(data(bL1,2),sig2,D2,alf2,r); xb2 = BS_1D(data(bL2,1),sig1,D1,alf1,r);
F = [E;xb1(2:end);xb2;zeros(length(bS),1)];
Wb = (.5*sig1^2*Idx^2*Axx(ind,inb)+.5*ru*sig1*sig2*Idx*Idy*Axy(ind,inb)+...
      .5*sig2^2*Idy^2*Ayy(ind,inb)+.5*ru*sig2*sig1*Idx*Idy*Axy(ind,inb)+...
      -(D1-r)*Idx*Ax(ind,inb)-(D2-r)*Idy*Ay(ind,inb))*F;
% Initial condition
Fin = E-(alf1*data(ind,1)+alf2*data(ind,2)); P0 = max(0,Fin);
t = 0:0.05:Tf; opt = odeset('RelTol',1e-10);
[T,P] = ode15s(@(t,P)odefun_2D(t,P,Fin,Wd,Wb,r),t,P0,opt);
runtime=toc
%Solotion in uniform grids using the solution of the adapted point
[xp,yp]=meshgrid(.9:.1:1.1,.9:.1:1.1); unid = [xp(:) yp(:)]; 
Papp = interpol(data(ind,:),unid,P(length(T),:)');
solution = [unid Papp]
% Plot the sparsity structure of the coefficient matrix
Q = amd(Wd);
figure(2); spy(Wd(Q,Q)); set(gca,'Fontsize',14);
title('Sparsity structure of AMD ordered W_d','FontSize',14);
%figures__________________________________________________________________
 figure(3); plot3(data(ind,1),data(ind,2),P(length(T),:),'*')
 xlim([0,4]); ylim([0,4]); zlim([0,1]); grid on;
 set(gca,'Fontsize',14); xlabel('S_1'); ylabel('S_2'); zlabel('P(S_1,S_2,0)');
 
 Pval=reshape(P(length(T),:)',N-2,N-2); 
S1=reshape(data(ind,1),N-2,N-2); S2=reshape(data(ind,2),N-2,N-2);
figure(4);
mesh(S1,S2,Pval); set(gca,'Fontsize',14);
zlim([0,1]); xlabel('S_1'); ylabel('S_2'); zlabel('P(S_1,S_2,0)');
   